# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.api.keyvalue.boto_base.base import MagnetoDBTestCase
from tempest.common.utils.data_utils import rand_name


class MagnetoDBScanTest(MagnetoDBTestCase):

    def setUp(self):
        super(MagnetoDBScanTest, self).setUp()
        self.tname = rand_name().replace('-', '')
        self.client.create_table(self.smoke_attrs,
                                 self.tname,
                                 self.smoke_schema,
                                 self.smoke_throughput)
        self.wait_for_table_active(self.tname)
        self.rck = self.addResourceCleanUp(self.client.delete_table,
                                           self.tname)

    def tearDown(self):
        super(MagnetoDBScanTest, self).tearDown()
        self.client.delete_table(self.tname)
        self.assertTrue(self.wait_for_table_deleted(self.tname))
        self.cancelResourceCleanUp(self.rck)

    def _verify_scan_response(self, response,
                              attributes_to_get=None, limit=None,
                              select=None, scan_filter=None,
                              exclusive_start_key=None,
                              return_consumed_capacity=None,
                              total_segments=None, segment=None):
        """
        The method implements common verifications of response format.

        More specific verifications should be implemented in the test-cases
        explicitly.
        """
        if return_consumed_capacity is None:
            self.assertNotIn('ConsumedCapacity', response)
        self.assertIn('ScannedCount', response)
        self.assertIn('Count', response)
        self.assertEqual(select == 'COUNT', 'Items' not in response)
        if select != 'COUNT':
            self.assertEqual(response['Count'], len(response['Items']))
        if limit is None:
            self.assertNotIn('LastEvaluatedKey', response)



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64

from tempest.api.keyvalue.rest_base.base import MagnetoDBTestCase
from tempest.common.utils.data_utils import rand_name
from tempest import exceptions
from tempest.test import attr


class MagnetoDBBatchWriteTest(MagnetoDBTestCase):

    def setUp(self):
        super(MagnetoDBBatchWriteTest, self).setUp()
        self.tname = rand_name().replace('-', '')

    @attr(type=['BWI-12'])
    def test_batch_write_put_empty_attr_name(self):
        self._create_test_table(self.build_x_attrs('S'),
                                self.tname,
                                self.smoke_schema,
                                wait_for_active=True)
        item = self.build_x_item('S', 'forum1', 'subject2',
                                 ('', 'N', '1000'))
        request_body = {'request_items': {self.tname: [{'put_request':
                                                        {'item': item}}]}}
        with self.assertRaises(exceptions.BadRequest):
            self.client.batch_write_item(request_body)

    @attr(type=['BWI-14'])
    def test_batch_write_put_b(self):
        self._create_test_table(self.build_x_attrs('S'),
                                self.tname,
                                self.smoke_schema,
                                wait_for_active=True)
        value = base64.b64encode('\xFF')
        item = self.build_x_item('S', 'forum1', 'subject2',
                                 ('message', 'B', value))
        request_body = {'request_items': {self.tname: [{'put_request':
                                                        {'item': item}}]}}
        headers, body = self.client.batch_write_item(request_body)
        self.assertIn('unprocessed_items', body)
        self.assertEqual(body['unprocessed_items'], {})
        key_conditions = {
            self.hashkey: {
                'attribute_value_list': [{'S': 'forum1'}],
                'comparison_operator': 'EQ'
            }
        }
        headers, body = self.client.query(self.tname,
                                          key_conditions=key_conditions,
                                          consistent_read=True)
        self.assertEqual(item, body['items'][0])

    @attr(type=['BWI-15'])
    def test_batch_write_put_bs(self):
        self._create_test_table(self.build_x_attrs('S'),
                                self.tname,
                                self.smoke_schema,
                                wait_for_active=True)
        value = base64.b64encode('\xFF')
        item = self.build_x_item('S', 'forum1', 'subject2',
                                 ('message', 'BS', [value, value]))
        request_body = {'request_items': {self.tname: [{'put_request':
                                                        {'item': item}}]}}
        headers, body = self.client.batch_write_item(request_body)
        self.assertIn('unprocessed_items', body)
        self.assertEqual(body['unprocessed_items'], {})
        key_conditions = {
            self.hashkey: {
                'attribute_value_list': [{'S': 'forum1'}],
                'comparison_operator': 'EQ'
            }
        }
        headers, body = self.client.query(self.tname,
                                          key_conditions=key_conditions,
                                          consistent_read=True)
        self.assertEqual(item, body['items'][0])

    @attr(type=['BWI-17'])
    def test_batch_write_put_n_empty_value(self):
        self._create_test_table(self.build_x_attrs('S'), self.tname,
                                self.smoke_schema,
                                wait_for_active=True)
        item = self.build_x_item('S', 'forum1', 'subject2',
                                 ('message', 'N', ''))
        request_body = {'request_items': {self.tname: [{'put_request':
                                                        {'item': item}}]}}
        with self.assertRaises(exceptions.BadRequest):
            self.client.batch_write_item(request_body)

    @attr(type=['BWI-42'])
    def test_batch_write_put_delete_same_item(self):
        self._create_test_table(self.smoke_attrs, self.tname,
                                self.smoke_schema,
                                wait_for_active=True)
        item = self.build_smoke_item('forum1', 'subject2',
                                     'message text', 'John', '10')
        key = {self.hashkey: {'S': 'forum1'}, self.rangekey: {'S': 'subject2'}}
        request_body = {'request_items': {self.tname: [{'put_request':
                                                        {'item': item}},
                                                       {'delete_request':
                                                        {'key': key}}]}}
        with self.assertRaises(exceptions.BadRequest):
            self.client.batch_write_item(request_body)

    @attr(type=['BWI-54_1'])
    def test_batch_write_too_short_tname(self):
        tname = 'qq'
        item = self.build_x_item('S', 'forum1', 'subject2',
                                 ('message', 'S', 'message text'))
        request_body = {'request_items': {tname: [{'put_request':
                                                   {'item': item}}]}}

        with self.assertRaises(exceptions.BadRequest):
            self.client.batch_write_item(request_body)

    @attr(type=['BWI-54_2'])
    def test_batch_write_too_long_tname(self):
        tname = 'q' * 256
        item = self.build_x_item('S', 'forum1', 'subject2',
                                 ('message', 'S', 'message text'))
        request_body = {'request_items': {tname: [{'put_request':
                                                   {'item': item}}]}}

        with self.assertRaises(exceptions.BadRequest):
            self.client.batch_write_item(request_body)

    @attr(type=['BWI-54_3'])
    def test_batch_write_non_existent_table(self):
        tname = 'non_existent_table'
        item = self.build_x_item('S', 'forum1', 'subject2',
                                 ('message', 'S', 'message text'))
        request_body = {'request_items': {tname: [{'put_request':
                                                   {'item': item}}]}}

        with self.assertRaises(exceptions.NotFound):
            self.client.batch_write_item(request_body)



# Copyright 2013 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from magnetodb.common import PROJECT_NAME, setup_global_env


def get_root_source(start_search_path, root_dir_name):
    cur_path = os.path.abspath(start_search_path)
    while (os.path.basename(cur_path) != root_dir_name):
        new_cur_path = os.path.dirname(cur_path)
        if (not new_cur_path) or (new_cur_path == cur_path):
            raise RuntimeError("Can't find project root directory.")
        cur_path = new_cur_path
    return os.path.dirname(cur_path)

PROJECT_ROOT_DIR = get_root_source(__file__, PROJECT_NAME)

setup_global_env(program="magnetodb-api", args=[
    "--config-file", PROJECT_ROOT_DIR+"/etc/magnetodb-api-test.conf"
])



# Copyright 2013 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from magnetodb.api.amz.dynamodb.action import DynamoDBAction
from magnetodb.api.amz.dynamodb import parser

from magnetodb import storage
from magnetodb.storage import models

from magnetodb.common import exception


class DeleteItemDynamoDBAction(DynamoDBAction):
    schema = {
        "required": [parser.Props.KEY,
                     parser.Props.TABLE_NAME],
        "properties": {
            parser.Props.EXPECTED: {
                "type": "object",
                "patternProperties": {
                    parser.ATTRIBUTE_NAME_PATTERN: {
                        "oneOf": [
                            {
                                "type": "object",
                                "required": [parser.Props.EXISTS],
                                "properties": {
                                    parser.Props.EXISTS: {
                                        "type": "boolean",
                                    },
                                }
                            },
                            {
                                "type": "object",
                                "required": [parser.Props.VALUE],
                                "properties": {
                                    parser.Props.VALUE:
                                        parser.Types.ITEM_VALUE,
                                }
                            },
                        ]
                    }
                }
            },

            parser.Props.KEY: {
                "type": "object",
                "patternProperties": {
                    parser.ATTRIBUTE_NAME_PATTERN: parser.Types.ITEM_VALUE
                }
            },

            parser.Props.RETURN_CONSUMED_CAPACITY: (
                parser.Types.RETURN_CONSUMED_CAPACITY
            ),
            parser.Props.RETURN_ITEM_COLLECTION_METRICS: (
                parser.Types.RETURN_ITEM_COLLECTION_METRICS
            ),
            parser.Props.RETURN_VALUES: {
                "type": "string",
                "enum": [parser.Values.RETURN_VALUES_NONE,
                         parser.Values.RETURN_VALUES_ALL_OLD]
            },
            parser.Props.TABLE_NAME: parser.Types.TABLE_NAME
        }
    }

    def __call__(self):
        try:
            table_name = self.action_params.get(parser.Props.TABLE_NAME, None)

            # parse expected item conditions
            expected_item_conditions = (
                parser.Parser.parse_expected_attribute_conditions(
                    self.action_params.get(parser.Props.EXPECTED, {})
                )
            )

            # parse item
            key_attributes = parser.Parser.parse_item_attributes(
                self.action_params[parser.Props.KEY]
            )

            # parse return_values param
            return_values = self.action_params.get(
                parser.Props.RETURN_VALUES, parser.Values.RETURN_VALUES_NONE
            )

            # parse return_item_collection_metrics
            return_item_collection_metrics = self.action_params.get(
                parser.Props.RETURN_ITEM_COLLECTION_METRICS,
                parser.Values.RETURN_ITEM_COLLECTION_METRICS_NONE
            )

            return_consumed_capacity = self.action_params.get(
                parser.Props.RETURN_CONSUMED_CAPACITY,
                parser.Values.RETURN_CONSUMED_CAPACITY_NONE
            )
        except Exception:
            raise exception.ValidationException()

        try:
            # put item
            result = storage.delete_item(
                self.context,
                models.DeleteItemRequest(table_name, key_attributes),
                expected_condition_map=expected_item_conditions)
        except exception.AWSErrorResponseException as e:
            raise e
        except Exception:
            raise exception.AWSErrorResponseException()

        if not result:
            raise exception.AWSErrorResponseException()

        # format response
        response = {}

        try:
            if return_values != parser.Values.RETURN_VALUES_NONE:
                # TODO(dukhlov):
                # It is needed to return all deleted item attributes
                #
                response[parser.Props.ATTRIBUTES] = (
                    parser.Parser.format_item_attributes(key_attributes)
                )

            if (return_item_collection_metrics !=
                    parser.Values.RETURN_ITEM_COLLECTION_METRICS_NONE):
                response[parser.Props.ITEM_COLLECTION_METRICS] = {
                    parser.Props.ITEM_COLLECTION_KEY: {
                        parser.Parser.format_item_attributes(
                            models.AttributeValue(models.ATTRIBUTE_TYPE_STRING,
                                                  "key")
                        )
                    },
                    parser.Props.SIZE_ESTIMATED_RANGE_GB: [0]
                }

            if (return_consumed_capacity !=
                    parser.Values.RETURN_CONSUMED_CAPACITY_NONE):
                response[parser.Props.CONSUMED_CAPACITY] = (
                    parser.Parser.format_consumed_capacity(
                        return_consumed_capacity, None
                    )
                )

            return response
        except Exception:
            raise exception.AWSErrorResponseException()



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import httplib
import json
import mock

from magnetodb.tests.unittests.api.openstack.v1 import test_base_testcase


class BatchWriteItemTestCase(test_base_testcase.APITestCase):
    """The test for batch_write_item method for v1 ReST API."""

    @mock.patch('magnetodb.storage.execute_write_batch', return_value=[])
    def test_batch_write_item(self, mock_execute_write_batch):
        headers = {'Content-Type': 'application/json',
                   'Accept': 'application/json'}

        conn = httplib.HTTPConnection('localhost:8080')
        url = '/v1/fake_project_id/data/batch_write_item'
        body = """
            {
                "request_items": {
                    "Forum": [
                        {
                            "put_request": {
                                "item": {
                                    "Name": {"S": "MagnetoDB"},
                                    "Category": {"S": "OpenStack KVaaS"}
                                }
                            }
                        },
                        {
                            "put_request": {
                                "item": {
                                    "Name": {"S": "Nova"},
                                    "Category": {"S": "OpenStack Core"}
                                }
                            }
                        },
                        {
                            "put_request": {
                                "item": {
                                    "Name": {"S": "KeyStone"},
                                    "Category": {"S": "OpenStack Core"}
                                }
                            }
                        },
                        {
                            "delete_request": {
                                "key": {
                                    "Name": {"S": "Cinder"},
                                    "Category": {"S": "OpenStack Core"}
                                }
                            }
                        }
                    ]
                }
            }
        """
        conn.request("POST", url, headers=headers, body=body)

        response = conn.getresponse()

        json_response = response.read()
        response_payload = json.loads(json_response)

        self.assertTrue(mock_execute_write_batch.called)
        self.assertEqual({'unprocessed_items': {}}, response_payload)

    def test_batch_write_item_malformed(self):
        self.maxDiff = None
        headers = {'Content-Type': 'application/json',
                   'Accept': 'application/json'}

        conn = httplib.HTTPConnection('localhost:8080')
        url = '/v1/fake_project_id/data/batch_write_item'
        body = '{"foo": "bar"}'
        conn.request("POST", url, headers=headers, body=body)

        response = conn.getresponse()

        json_response = response.read()
        response_payload = json.loads(json_response)

        self.assertEqual(400, response.status)

        expected_message = "'request_items' is a required property"
        expected_type = 'ValidationError'

        self.assertEqual(expected_message,
                         response_payload['error']['message'])
        self.assertEqual(expected_type,
                         response_payload['error']['type'])



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import httplib
import unittest

from magnetodb.tests.fake import magnetodb_api_fake
import mock


class APITest(unittest.TestCase):
    """
    The test for low level API calls
    """

    @classmethod
    def setUpClass(cls):
        magnetodb_api_fake.run_fake_magnetodb_api()

    @classmethod
    def tearDownClass(cls):
        magnetodb_api_fake.stop_fake_magnetodb_api()

    @mock.patch(
        "magnetodb.common.middleware.fault.FaultWrapper.process_request"
    )
    def test_connection_header(self, magnetodb_app_mock):
        magnetodb_app_mock.return_value = "{}"

        headers = {}

        conn = httplib.HTTPConnection('localhost:8080')
        conn.request("POST", "/", body='{}', headers=headers)

        response = conn.getresponse()

        self.assertEqual(response.getheader('Connection'), None)

        headers = {"Connection": "close"}

        conn = httplib.HTTPConnection('localhost:8080')
        conn.request("POST", "/", body='{}', headers=headers)

        response = conn.getresponse()

        self.assertEqual(response.getheader('Connection'), 'close')

        self.assertEqual(magnetodb_app_mock.call_count, 2)



# Copyright 2013 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json

import unittest
import uuid
import binascii
from magnetodb.common import exception

from cassandra import cluster
from cassandra import query
from magnetodb.common.cassandra.cluster_handler import ClusterHandler
from magnetodb.storage import models
from magnetodb.storage.driver.cassandra import (
    cassandra_impl,
    SYSTEM_COLUMN_INDEX_NAME,
    SYSTEM_COLUMN_INDEX_VALUE_STRING,
    SYSTEM_COLUMN_INDEX_VALUE_NUMBER,
    SYSTEM_COLUMN_INDEX_VALUE_BLOB,
    USER_PREFIX,
    SYSTEM_COLUMN_EXTRA_ATTR_DATA,
    SYSTEM_COLUMN_EXTRA_ATTR_TYPES,
    SYSTEM_COLUMN_ATTR_EXIST
)
from magnetodb.storage.manager.simple_impl import SimpleStorageManager
from magnetodb.storage.table_info_repo.cassandra_impl import (
    CassandraTableInfoRepository
)

TEST_CONNECTION = {
    'contact_points': ("localhost",),
    'control_connection_timeout': 60
}


class FakeContext(object):
    def __init__(self, tenant):
        self.tenant = tenant


class TestCassandraBase(unittest.TestCase):
    TENANT_PER_TEST_METHOD = "test"
    TENANT_PER_TEST_CLASS = "class"

    _tenant_scope = TENANT_PER_TEST_CLASS

    test_data_keys = {
        'id': ('decimal', '1', 1),
        'range': ('text', "'1'", '1')
    }

    test_table_schema = models.TableSchema(
        {
            "id": models.ATTRIBUTE_TYPE_NUMBER,
            "range": models.ATTRIBUTE_TYPE_STRING,
            "indexed": models.ATTRIBUTE_TYPE_STRING,
            "str": models.ATTRIBUTE_TYPE_STRING,
            "numbr": models.ATTRIBUTE_TYPE_NUMBER,
            "blb": models.ATTRIBUTE_TYPE_BLOB,
            "set_number": models.ATTRIBUTE_TYPE_NUMBER_SET,
            "set_string": models.ATTRIBUTE_TYPE_STRING_SET,
            "set_blob": models.ATTRIBUTE_TYPE_BLOB_SET
        },
        ["id", "range"]
    )

    test_table_schema_with_index = models.TableSchema(
        {
            "id": models.ATTRIBUTE_TYPE_NUMBER,
            "range": models.ATTRIBUTE_TYPE_STRING,
            "indexed": models.ATTRIBUTE_TYPE_STRING,
            "str": models.ATTRIBUTE_TYPE_STRING,
            "numbr": models.ATTRIBUTE_TYPE_NUMBER,
            "blb": models.ATTRIBUTE_TYPE_BLOB,
            "set_number": models.ATTRIBUTE_TYPE_NUMBER_SET,
            "set_string": models.ATTRIBUTE_TYPE_STRING_SET,
            "set_blob": models.ATTRIBUTE_TYPE_BLOB_SET
        },
        ["id", "range"],
        {
            "index": models.IndexDefinition("indexed")
        }
    )

    test_data_predefined_fields = {
        'indexed': ('text', "'ind'", 'ind'),
        'str': ('text', "'str'", 'str'),
        'numbr': ('decimal', '1', 1),
        'blb': ('blob', '0x{}'.format(binascii.hexlify('blob')), 'blob'),
        'set_number': ('set<decimal>', '{1,2,3}', {1, 2, 3}),
        'set_string': ('set<text>', "{'a','b','c'}", {'a', 'b', 'c'}),
        'set_blob': (
            'set<blob>', '{{0x{}, 0x{}}}'.format(
                binascii.hexlify('blob1'), binascii.hexlify('blob2')),
            {'blob1', 'blob2'}
        )
    }

    test_data_system_fields = {
        SYSTEM_COLUMN_EXTRA_ATTR_DATA: 'map<text,blob>',
        SYSTEM_COLUMN_EXTRA_ATTR_TYPES: 'map<text,text>',
        SYSTEM_COLUMN_ATTR_EXIST: 'set<text>'
    }

    test_data_dynamic_fields = {
        'fnum': ('decimal', binascii.hexlify(json.dumps('1')), 1),
        'fstr': ('text', binascii.hexlify(json.dumps('fstr')), 'fstr'),
        'fblb': ('blob', binascii.hexlify(json.dumps('fblob')), 'fblob'),
        'fsnum': (
            'set<decimal>', binascii.hexlify(json.dumps(['1', '2', '3'])),
            {1, 2, 3}
        ),
        'fsstr': (
            'set<text>', binascii.hexlify(
                json.dumps(['fa', 'fb', 'fc'])), {'fa', 'fb', 'fc'}
        ),
        'fsblob': (
            'set<blob>', binascii.hexlify(json.dumps(['fblob1', 'fblob2'])),
            {'fblob1', 'fblob2'}
        )
    }

    C2S_TYPES = cassandra_impl.CASSANDRA_TO_STORAGE_TYPES

    @classmethod
    def setUpClass(cls):
        super(TestCassandraBase, cls).setUpClass()

        cls.CLUSTER = cluster.Cluster(**TEST_CONNECTION)
        cluster_hadler = ClusterHandler(cls.CLUSTER, query_timeout=300)
        table_info_repo = CassandraTableInfoRepository(cluster_hadler)
        storage_driver = cassandra_impl.CassandraStorageDriver(
            cluster_hadler, table_info_repo
        )
        cls.CASANDRA_STORAGE_IMPL = SimpleStorageManager(storage_driver,
                                                         table_info_repo)

        cls.SESSION = cls.CLUSTER.connect()
        cls.SESSION.row_factory = query.dict_factory
        cls.SESSION.default_timeout = 300

        if cls._tenant_scope == cls.TENANT_PER_TEST_CLASS:
            cls.tenant = cls._get_unique_name()
            cls._create_tenant(cls.tenant)

        cls.expected_data = {
            name: models.AttributeValue(cls.C2S_TYPES[typ], val)
            for name, (typ, _, val)
            in dict(cls.test_data_keys.items() +
                    cls.test_data_dynamic_fields.items() +
                    cls.test_data_predefined_fields.items()).iteritems()
        }

    @classmethod
    def tearDownClass(cls):
        super(TestCassandraBase, cls).tearDownClass()
        if cls._tenant_scope == cls.TENANT_PER_TEST_CLASS:
            cls._drop_tenant(cls.tenant)

    def setUp(self):
        if self._tenant_scope == self.TENANT_PER_TEST_METHOD:
            self.keyspace = self._get_unique_name()
            self._create_tenant(self.tenant)

        self.context = FakeContext(self.tenant)

        self.table_name = self._get_unique_name()

    def tearDown(self):
        if self._tenant_scope == self.TENANT_PER_TEST_METHOD:
            self._drop_tenant(self.tenant)

    @staticmethod
    def _get_unique_name():
        name = str(uuid.uuid4())
        return 'test' + filter(lambda x: x != '-', name)[:28]

    @classmethod
    def _create_tenant(cls, tenant):
        query = "CREATE KEYSPACE {}{} WITH replication".format(
            USER_PREFIX, tenant)
        query += " = {'class':'SimpleStrategy', 'replication_factor':1}"

        cls.SESSION.execute(query)

    @classmethod
    def _drop_tenant(cls, tenant):
        query = ("DROP KEYSPACE {}{}".format(USER_PREFIX, tenant))

        cls.SESSION.execute(query)

    def _get_table_names(self, tenant=None):

        query = (
            "SELECT name FROM magnetodb.table_info WHERE tenant='{}'"
        ).format(
            tenant or self.tenant
        )

        result = self.SESSION.execute(query)
        return [item["name"] for item in result]

    def _create_table(self, tenant=None, table_name=None, indexed=True):
        tenant = tenant or self.tenant
        table_name = table_name or self.table_name

        internal_table_name = USER_PREFIX + self._get_unique_name()
        keyspace = USER_PREFIX + tenant

        query = (
            "INSERT INTO magnetodb.table_info (tenant, name, exists, "
            '"schema", status, internal_name) '
            "VALUES('{}', '{}', 1, '{}', 'active', '{}') IF NOT EXISTS"
        ).format(
            tenant, table_name,
            self.test_table_schema_with_index.to_json() if indexed else
            self.test_table_schema.to_json(),
            internal_table_name
        )
        result = self.SESSION.execute(query)
        self.assertTrue(result[0]['[applied]'])

        query = "CREATE TABLE {}.{} (".format(keyspace, internal_table_name)

        for name, field in self.test_data_keys.iteritems():
            typ, _, _ = field
            query += '{}{} {},'.format(USER_PREFIX, name, typ)

        for name, field in self.test_data_predefined_fields.iteritems():
            typ, _, _ = field
            query += '{}{} {},'.format(USER_PREFIX, name, typ)

        for name, field in self.test_data_system_fields.iteritems():
            query += '{} {},'.format(name, field)

        if indexed:
            query += (
                "{} text, {} text, {} decimal, {} blob,"
                " PRIMARY KEY({}id, {}, {}, {}, {}, {}range))".format(
                    SYSTEM_COLUMN_INDEX_NAME,
                    SYSTEM_COLUMN_INDEX_VALUE_STRING,
                    SYSTEM_COLUMN_INDEX_VALUE_NUMBER,
                    SYSTEM_COLUMN_INDEX_VALUE_BLOB,
                    USER_PREFIX,
                    SYSTEM_COLUMN_INDEX_NAME,
                    SYSTEM_COLUMN_INDEX_VALUE_STRING,
                    SYSTEM_COLUMN_INDEX_VALUE_NUMBER,
                    SYSTEM_COLUMN_INDEX_VALUE_BLOB,
                    USER_PREFIX
                )
            )
        else:
            query += " PRIMARY KEY({}id, {}range))".format(
                USER_PREFIX, USER_PREFIX
            )
        self.SESSION.execute(query)

    def _drop_table(self, tenant=None, table_name=None):
        query = (
            "SELECT internal_name FROM magnetodb.table_info "
            "WHERE tenant='{}' and name='{}'"
        ).format(
            tenant, table_name
        )
        result = self.SESSION.execute(query)
        internal_table_name = result[0]["internal_name"]
        query = "DROP TABLE IF EXISTS {}{}.{}".format(
            USER_PREFIX, tenant, internal_table_name
        )
        self.SESSION.execute(query)
        query = (
            "DELETE FROM magnetodb.table_info "
            "WHERE tenant='{}' and name='{}'".format(
                tenant, table_name
            )
        )
        self.SESSION.execute(query)

    def _select_all(self, tenant=None, table_name=None):
        tenant = tenant or self.tenant
        table_name = table_name or self.table_name
        query = (
            'SELECT internal_name, "schema" FROM magnetodb.table_info '
            "WHERE tenant='{}' and name='{}'"
        ).format(
            tenant, table_name
        )
        result = self.SESSION.execute(query)
        internal_table_name = result[0]["internal_name"]
        schema = models.ModelBase.from_json(result[0]["schema"])

        query = "SELECT * FROM {}{}.{}".format(
            USER_PREFIX, tenant, internal_table_name)

        if schema.index_def_map:
            query += " WHERE {}={}".format(
                SYSTEM_COLUMN_INDEX_NAME,
                cassandra_impl.ENCODED_DEFAULT_STRING_VALUE
            )

        query += " ALLOW FILTERING"

        return self.SESSION.execute(query)

    def _insert_data(self, id_value=1, range_value='1',
                     predefined_fields=None, dynamic_fields=None):
        query = (
            'SELECT internal_name, "schema" FROM magnetodb.table_info '
            "WHERE tenant='{}' and name='{}'"
        ).format(
            self.tenant, self.table_name
        )

        result = self.SESSION.execute(query)
        internal_table_name = result[0]["internal_name"]
        schema = models.ModelBase.from_json(result[0]["schema"])

        query = "UPDATE {}{}.{} SET ".format(
            USER_PREFIX, self.tenant, internal_table_name)

        predefined_fields = (
            predefined_fields or self.test_data_predefined_fields
        )

        dynamic_fields = dynamic_fields or self.test_data_dynamic_fields

        set_items = []
        for name, field in predefined_fields.iteritems():
            _, sval, _ = field
            set_items.append('{}{}={}'.format(USER_PREFIX, name, sval))

        for name, field in dynamic_fields.iteritems():
            typ, sval, _ = field
            set_items.append("{}['{}'] = 0x{}".format(
                SYSTEM_COLUMN_EXTRA_ATTR_DATA, name, sval))
            set_items.append("{}['{}'] ='{}'".format(
                SYSTEM_COLUMN_EXTRA_ATTR_TYPES, name, typ))

        for name, field in dict(self.test_data_keys.items() +
                                predefined_fields.items() +
                                dynamic_fields.items()).iteritems():
            typ, sval, _ = field

            set_items.append("{} = {} + {{'{}'}}".format(
                SYSTEM_COLUMN_ATTR_EXIST, SYSTEM_COLUMN_ATTR_EXIST,
                name,
            ))

        query += ",".join(set_items)

        query += " WHERE {}id = {} AND {}range='{}'".format(
            USER_PREFIX, id_value, USER_PREFIX, range_value)

        if schema.index_def_map:
            default_index_cond_params = (
                SYSTEM_COLUMN_INDEX_NAME,
                cassandra_impl.ENCODED_DEFAULT_STRING_VALUE,
                SYSTEM_COLUMN_INDEX_VALUE_BLOB,
                cassandra_impl.ENCODED_DEFAULT_BLOB_VALUE,
                SYSTEM_COLUMN_INDEX_VALUE_STRING,
                cassandra_impl.ENCODED_DEFAULT_STRING_VALUE,
                SYSTEM_COLUMN_INDEX_VALUE_NUMBER,
                cassandra_impl.ENCODED_DEFAULT_NUMBER_VALUE
            )

            queries = [
                query + (
                    " AND {}={} AND {}={} AND {}={} AND {}={}".format(
                        *default_index_cond_params
                    )
                )
            ]

            for index_name, index_def in schema.index_def_map.iteritems():
                params = list(default_index_cond_params)
                params[1] = '{}'.format(index_name)
                index_type = schema.attribute_type_map[
                    index_def.attribute_to_index]

                index_value = predefined_fields[
                    index_def.attribute_to_index
                ][1]
                if index_type == models.ATTRIBUTE_TYPE_BLOB:
                    params[3] = index_value
                elif index_type == models.ATTRIBUTE_TYPE_STRING:
                    params[5] = index_value
                elif index_type == models.ATTRIBUTE_TYPE_NUMBER:
                    params[7] = index_value
                else:
                    self.fail()

                queries.append(
                    query +
                    " AND {}='{}' AND {}={} AND {}={} AND {}={}".format(
                        *params
                    )
                )

            query = (
                "BEGIN UNLOGGED BATCH " + " ".join(queries) + " APPLY BATCH"
            )

        self.SESSION.execute(query)

    def _validate_data(self, data):

        self.assertDictEqual(self.expected_data, data)


class TestCassandraTableCrud(TestCassandraBase):

    def test_create_table(self):
        self.assertEqual([], self._get_table_names())

        attrs = {}
        for name, (typ, _, _) in self.test_data_keys.iteritems():
            attrs[name] = self.C2S_TYPES[typ]
        for name, (typ, _, _) in self.test_data_predefined_fields.iteritems():
            attrs[name] = self.C2S_TYPES[typ]

        index_def_map = {
            'index_name': models.IndexDefinition('indexed')
        }

        schema = models.TableSchema(attrs, ['id', 'range'],
                                    index_def_map)

        self.CASANDRA_STORAGE_IMPL.create_table(
            self.context, self.table_name, schema
        )

        self.assertEqual([self.table_name], self._get_table_names())

    def test_list_table(self):
        self.assertNotIn(self.table_name,
                         self.CASANDRA_STORAGE_IMPL.list_tables(self.context))

        self._create_table()

        self.assertIn(self.table_name,
                      self.CASANDRA_STORAGE_IMPL.list_tables(self.context))

    def test_describe_table(self):

        self._create_table(indexed=True)

        attrs = {}
        for name, (typ, _, _) in self.test_data_keys.iteritems():
            attrs[name] = self.C2S_TYPES[typ]
        for name, (typ, _, _) in self.test_data_predefined_fields.iteritems():
            attrs[name] = self.C2S_TYPES[typ]

        index_def_map = {
            'index': models.IndexDefinition('indexed')
        }

        schema = models.TableSchema(attrs, ['id', 'range'],
                                    index_def_map)

        desc = self.CASANDRA_STORAGE_IMPL.describe_table(
            self.context, self.table_name)

        self.assertEqual(schema, desc.schema)

    def test_delete_table(self):
        self._create_table()

        self.assertIn(self.table_name, self._get_table_names())

        self.CASANDRA_STORAGE_IMPL.delete_table(self.context, self.table_name)

        self.assertNotIn(self.table_name, self._get_table_names())


class TestCassandraDeleteItem(TestCassandraBase):

    def test_delete_item_where(self):
        self._create_table(indexed=True)
        self._insert_data()
        self._insert_data(range_value=2)

        del_req = models.DeleteItemRequest(
            self.table_name,
            {
                'id': models.AttributeValue.number(1),
                'range': models.AttributeValue.str('1')
            }
        )

        self.CASANDRA_STORAGE_IMPL.delete_item(
            self.context, del_req)

        all = self._select_all()

        self.assertEqual(1, len(all))
        self.assertEqual('2', all[0]['user_range'])

    def test_delete_item_where_negative(self):
        self._create_table(indexed=True)
        self._insert_data()

        del_req = models.DeleteItemRequest(
            self.table_name,
            {'id': models.AttributeValue.number(1),
             'range': models.AttributeValue.str('2')})

        self.CASANDRA_STORAGE_IMPL.delete_item(
            self.context, del_req)

        all = self._select_all()

        self.assertEqual(1, len(all))
        self.assertEqual(1, all[0]['user_id'])

    @unittest.skip("Skipped due to Cassandra 2.0.6 bug"
                   "(https://issues.apache.org/jira/browse/CASSANDRA-6914)")
    def test_delete_item_if_exists(self):
        self._create_table(indexed=True)
        self._insert_data()

        all = self._select_all()

        self.assertEqual(1, len(all))
        self.assertEqual(1, all[0]["user_id"])

        expected = {'str': [models.ExpectedCondition.exists()]}

        del_req = models.DeleteItemRequest(
            self.table_name,
            {'id': models.AttributeValue.number(1),
             'range': models.AttributeValue.str('1')})

        self.CASANDRA_STORAGE_IMPL.delete_item(
            self.context, del_req, expected
        )

        all = self._select_all()

        self.assertEqual(0, len(all))

    def test_delete_item_if_exists_negative(self):
        self._create_table(indexed=True)

        self._insert_data()

        all = self._select_all()

        self.assertEqual(1, len(all))
        self.assertEqual(1, all[0]["user_id"])

        expected = {
            'not_existed_attr_name': [models.ExpectedCondition.exists()]
        }

        del_req = models.DeleteItemRequest(
            self.table_name,
            {'id': models.AttributeValue.number(1),
             'range': models.AttributeValue.str('1')})

        self.assertRaises(exception.ConditionalCheckFailedException,
                          self.CASANDRA_STORAGE_IMPL.delete_item,
                          self.context, del_req, expected)

        all = self._select_all()

        self.assertEqual(1, len(all))
        self.assertEqual(1, all[0]["user_id"])

    @unittest.skip("Skipped due to Cassandra 2.0.6 bug"
                   "(https://issues.apache.org/jira/browse/CASSANDRA-6914)")
    def test_delete_item_if_not_exists(self):
        self._create_table(indexed=True)

        self._insert_data()

        all = self._select_all()

        self.assertEqual(1, len(all))
        self.assertEqual(1, all[0]["user_id"])

        expected = {
            'not_existed_attr_name': [models.ExpectedCondition.not_exists()]
        }

        del_req = models.DeleteItemRequest(
            self.table_name,
            {'id': models.AttributeValue.number(1),
             'range': models.AttributeValue.str('1')})

        result = self.CASANDRA_STORAGE_IMPL.delete_item(
            self.context, del_req, expected)

        self.assertTrue(result)

        all = self._select_all()

        self.assertEqual(0, len(all))

    def test_delete_item_if_not_exists_negative(self):
        self._create_table(indexed=True)
        self._insert_data()

        all = self._select_all()

        self.assertEqual(1, len(all))
        self.assertEqual(1, all[0]["user_id"])

        expected = {'str': [models.ExpectedCondition.not_exists()]}

        del_req = models.DeleteItemRequest(
            self.table_name,
            {'id': models.AttributeValue.number(1),
             'range': models.AttributeValue.str('1')})

        self.assertRaises(exception.ConditionalCheckFailedException,
                          self.CASANDRA_STORAGE_IMPL.delete_item,
                          self.context, del_req, expected)

        all = self._select_all()

        self.assertEqual(1, len(all))
        self.assertEqual(1, all[0]["user_id"])


class TestCassandraSelectItem(TestCassandraBase):

    def test_select_item(self):
        self._create_table(indexed=True)

        self._insert_data()

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_select_item_no_condition(self):
        self._create_table(indexed=True)

        self._insert_data()

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name)

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_select_item_attr(self):
        self._create_table(indexed=True)

        self._insert_data()

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond,
            models.SelectType.specified_attributes(['fstr'])
        )

        self.assertEqual(1, result.count)
        self.assertEqual(
            {'fstr': models.AttributeValue(
                models.ATTRIBUTE_TYPE_STRING, 'fstr')},
            result.items[0])

    def test_select_item_negative(self):
        self._create_table(indexed=True)

        self._insert_data()

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('2'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(0, result.count)

    def test_select_item_less(self):
        self._create_table(indexed=True)

        self._insert_data()

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.lt(models.AttributeValue.str('2'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_select_item_less_negative(self):
        self._create_table(indexed=True)

        self._insert_data()

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.lt(models.AttributeValue.str('1'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(0, result.count)

    def test_select_item_less_eq(self):
        self._create_table(indexed=True)

        self._insert_data()

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.le(models.AttributeValue.str('1'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_select_item_less_eq_negative(self):
        self._create_table(indexed=True)

        self._insert_data()

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.le(models.AttributeValue.str('0'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(0, result.count)

    def test_select_item_greater(self):
        self._create_table(indexed=True)

        self._insert_data()

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.gt(models.AttributeValue.str('0'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_select_item_greater_negative(self):
        self._create_table(indexed=True)

        self._insert_data()

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.gt(models.AttributeValue.str('1'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(0, result.count)

    def test_select_item_greater_eq(self):
        self._create_table(indexed=True)

        self._insert_data()

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.ge(models.AttributeValue.str('1'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_select_item_greater_eq_negative(self):
        self._create_table(indexed=True)

        self._insert_data()

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.ge(models.AttributeValue.str('2'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(0, result.count)

    def test_select_item_indexed(self):
        self._create_table(indexed=True)

        self._insert_data()

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'indexed': [
                models.IndexedCondition.le(models.AttributeValue.str('ind'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond, index_name="index")

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_select_item_indexed_negative(self):
        self._create_table(indexed=True)

        self._insert_data()

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'indexed': [
                models.IndexedCondition.lt(models.AttributeValue.str('ind'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond, index_name="index")

        self.assertEqual(0, result.count)

    def test_select_item_between(self):
        self._create_table(indexed=True)

        self._insert_data(range_value='0')
        self._insert_data(range_value='1')
        self._insert_data(range_value='3')

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.ge(models.AttributeValue.str('1')),
                models.IndexedCondition.le(models.AttributeValue.str('2'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_select_item_between2(self):
        self._create_table(indexed=True)

        self._insert_data(range_value='-1')
        self._insert_data(range_value='1')
        self._insert_data(range_value='2')

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.ge(models.AttributeValue.str('0')),
                models.IndexedCondition.le(models.AttributeValue.str('1'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_select_item_between_negative(self):
        self._create_table(indexed=True)

        self._insert_data(range_value='1')
        self._insert_data(range_value='4')

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.ge(models.AttributeValue.str('2')),
                models.IndexedCondition.le(models.AttributeValue.str('3'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(0, result.count)

    def test_select_item_begins_with(self):
        self._create_table(indexed=True)

        self._insert_data(range_value='0')
        self._insert_data(range_value='1')
        self._insert_data(range_value='2')

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.ge(models.AttributeValue.str('1')),
                models.IndexedCondition.lt(models.AttributeValue.str(
                    '1'[:-1] + chr(ord('1'[-1]) + 1))
                )
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_select_item_begins_with2(self):
        self._create_table(indexed=True)

        self._insert_data(range_value='0')
        self._insert_data(range_value='11')
        self._insert_data(range_value='2')

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.ge(models.AttributeValue.str('1')),
                models.IndexedCondition.lt(models.AttributeValue.str(
                    '1'[:-1] + chr(ord('1'[-1]) + 1))
                )
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(1, result.count)
        self.assertIn('11', result.items[0]['range'].value)

    def test_select_item_begins_with_negative(self):
        self._create_table(indexed=True)

        self._insert_data(range_value='0')
        self._insert_data(range_value='2')

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.ge(models.AttributeValue.str('1')),
                models.IndexedCondition.lt(models.AttributeValue.str(
                    '1'[:-1] + chr(ord('1'[-1]) + 1))
                )
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(0, result.count)

    def test_select_with_limit(self):
        self._create_table(indexed=True)
        self._insert_data()
        self._insert_data(range_value='2')

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.ge(
                    models.AttributeValue.str('1')
                )
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond)

        self.assertEqual(2, result.count)

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond, limit=1)

        self.assertEqual(1, result.count)

    def test_select_count(self):
        self._create_table(indexed=True)

        self._insert_data()

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond,
            models.SelectType.count())

        self.assertEqual(1, result.count)

    def test_select_item_exclusive_key(self):
        self._create_table(indexed=True)
        self._insert_data()

        exclusive_start_key = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('0')
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name,
            exclusive_start_key=exclusive_start_key)

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_select_item_exclusive_key2(self):
        self._create_table(indexed=True)

        self._insert_data()
        self._insert_data(id_value=2)

        exclusive_start_key = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('0')
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, limit=1,
            exclusive_start_key=exclusive_start_key)

        exclusive_start_key2 = {
            'id': models.AttributeValue.number(2),
            'range': models.AttributeValue.str('0')
        }

        result2 = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, limit=1,
            exclusive_start_key=exclusive_start_key2)

        self.assertTrue(result.count == 1 or result2.count == 1)

    def test_select_item_exclusive_key_negative(self):
        self._create_table(indexed=True)

        self._insert_data()

        exclusive_start_key = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1')
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name,
            exclusive_start_key=exclusive_start_key)

        self.assertEqual(0, result.count)

    def test_select_item_exclusive_key_with_range(self):
        self._create_table(indexed=True)

        self._insert_data(range_value='1')
        self._insert_data(range_value='2')
        self._insert_data(range_value='3')
        self._insert_data(range_value='4')
        self._insert_data(range_value='5')

        indexed_cond = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [
                models.IndexedCondition.gt(models.AttributeValue.str('1'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond,
            limit=2)

        self.assertEqual(2, result.count)
        self.assertEqual('2', result.items[0]['range'].value)
        self.assertEqual('3', result.items[1]['range'].value)

        last_eval_key = result.last_evaluated_key

        self.assertIsNotNone(last_eval_key)

        result2 = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, indexed_cond,
            exclusive_start_key=last_eval_key)

        self.assertEqual(2, result2.count)
        self.assertEqual('4', result2.items[0]['range'].value)
        self.assertEqual('5', result2.items[1]['range'].value)


class TestCassandraUpdateItem(TestCassandraBase):
    def test_update_item_put_str(self):
        self._create_table(indexed=True)
        self._insert_data()

        keys = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1')
        }

        actions = {
            'str': models.UpdateItemAction(
                models.UpdateItemAction.UPDATE_ACTION_PUT,
                models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, 'new')),
        }

        self.CASANDRA_STORAGE_IMPL.update_item(
            self.context, self.table_name, keys, actions)

        expected = self.expected_data.copy()
        expected['str'] = models.AttributeValue(
            models.ATTRIBUTE_TYPE_STRING, 'new')

        keys_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, keys_condition)

        self.assertEqual([expected], result.items)

    def test_update_item_put_number(self):
        self._create_table(indexed=True)
        self._insert_data()

        keys = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1')
        }

        actions = {
            'numbr': models.UpdateItemAction(
                models.UpdateItemAction.UPDATE_ACTION_PUT,
                models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 42)),
        }

        self.CASANDRA_STORAGE_IMPL.update_item(
            self.context, self.table_name, keys, actions)

        expected = self.expected_data.copy()

        expected['numbr'] = models.AttributeValue(
            models.ATTRIBUTE_TYPE_NUMBER, 42)

        keys_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, keys_condition)

        self.assertEqual([expected], result.items)

    def test_update_item_put_blob(self):
        self._create_table(indexed=True)
        self._insert_data()

        keys = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1')
        }

        actions = {
            'blb': models.UpdateItemAction(
                models.UpdateItemAction.UPDATE_ACTION_PUT,
                models.AttributeValue(models.ATTRIBUTE_TYPE_BLOB, 'new')),
        }

        self.CASANDRA_STORAGE_IMPL.update_item(
            self.context, self.table_name, keys, actions)

        expected = self.expected_data.copy()

        expected['blb'] = models.AttributeValue(
            models.ATTRIBUTE_TYPE_BLOB, 'new')

        keys_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, keys_condition)

        self.assertEqual([expected], result.items)

    def test_update_item_put_set_str(self):
        self._create_table(indexed=True)
        self._insert_data()

        keys = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1')
        }

        actions = {
            'set_string': models.UpdateItemAction(
                models.UpdateItemAction.UPDATE_ACTION_PUT,
                models.AttributeValue(
                    models.ATTRIBUTE_TYPE_STRING_SET, {'new'})),
        }

        self.CASANDRA_STORAGE_IMPL.update_item(
            self.context, self.table_name, keys, actions)

        expected = self.expected_data.copy()

        expected['set_string'] = models.AttributeValue(
            models.ATTRIBUTE_TYPE_STRING_SET, {'new'})

        keys_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, keys_condition)

        self.assertEqual([expected], result.items)

    def test_update_item_put_set_number(self):
        self._create_table(indexed=True)
        self._insert_data()

        keys = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1')
        }

        actions = {
            'set_number': models.UpdateItemAction(
                models.UpdateItemAction.UPDATE_ACTION_PUT,
                models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER_SET, {42})),
        }

        self.CASANDRA_STORAGE_IMPL.update_item(
            self.context, self.table_name, keys, actions)

        expected = self.expected_data.copy()

        expected['set_number'] = models.AttributeValue(
            models.ATTRIBUTE_TYPE_NUMBER_SET, {42})

        keys_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, keys_condition)

        self.assertEqual([expected], result.items)

    def test_update_item_put_set_blob(self):
        self._create_table(indexed=True)
        self._insert_data()

        keys = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1')
        }

        actions = {
            'set_blob': models.UpdateItemAction(
                models.UpdateItemAction.UPDATE_ACTION_PUT,
                models.AttributeValue(
                    models.ATTRIBUTE_TYPE_BLOB_SET, {'new'})),
        }

        self.CASANDRA_STORAGE_IMPL.update_item(
            self.context, self.table_name, keys, actions)

        expected = self.expected_data.copy()

        expected['set_blob'] = models.AttributeValue(
            models.ATTRIBUTE_TYPE_BLOB_SET, {'new'})

        keys_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, keys_condition)

        self.assertEqual([expected], result.items)

    def test_update_item_put_dynamic_str(self):
        self._create_table(indexed=True)
        self._insert_data()

        keys = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1')
        }

        actions = {
            'fstr': models.UpdateItemAction(
                models.UpdateItemAction.UPDATE_ACTION_PUT,
                models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, 'new')),
        }

        self.CASANDRA_STORAGE_IMPL.update_item(
            self.context, self.table_name, keys, actions)

        expected = self.expected_data.copy()

        expected['fstr'] = models.AttributeValue(
            models.ATTRIBUTE_TYPE_STRING, 'new')

        keys_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, keys_condition)

        self.assertEqual([expected], result.items)

    def test_update_item_put_dynamic_number(self):
        self._create_table(indexed=True)
        self._insert_data()

        keys = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1')
        }

        actions = {
            'fnum': models.UpdateItemAction(
                models.UpdateItemAction.UPDATE_ACTION_PUT,
                models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 42)),
        }

        self.CASANDRA_STORAGE_IMPL.update_item(
            self.context, self.table_name, keys, actions)

        expected = self.expected_data.copy()

        expected['fnum'] = models.AttributeValue(
            models.ATTRIBUTE_TYPE_NUMBER, 42)

        keys_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, keys_condition)

        self.assertEqual([expected], result.items)

    def test_update_item_put_dynamic_blob(self):
        self._create_table(indexed=True)
        self._insert_data()

        keys = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1')
        }

        actions = {
            'fblb': models.UpdateItemAction(
                models.UpdateItemAction.UPDATE_ACTION_PUT,
                models.AttributeValue(models.ATTRIBUTE_TYPE_BLOB, 'new')),
        }

        self.CASANDRA_STORAGE_IMPL.update_item(
            self.context, self.table_name, keys, actions)

        expected = self.expected_data.copy()

        expected['fblb'] = models.AttributeValue(
            models.ATTRIBUTE_TYPE_BLOB, 'new')

        keys_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, keys_condition)

        self.assertEqual([expected], result.items)

    def test_update_item_put_dynamic_set_str(self):
        self._create_table(indexed=True)
        self._insert_data()

        keys = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1')
        }

        actions = {
            'fsstr': models.UpdateItemAction(
                models.UpdateItemAction.UPDATE_ACTION_PUT,
                models.AttributeValue(
                    models.ATTRIBUTE_TYPE_STRING_SET, {'new1', 'new2'})),
        }

        self.CASANDRA_STORAGE_IMPL.update_item(
            self.context, self.table_name, keys, actions)

        expected = self.expected_data.copy()

        expected['fsstr'] = models.AttributeValue(
            models.ATTRIBUTE_TYPE_STRING_SET, {'new1', 'new2'})

        keys_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, keys_condition)

        self.assertEqual([expected], result.items)

    def test_update_item_put_dynamic_set_number(self):
        self._create_table(indexed=True)
        self._insert_data()

        keys = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1')
        }

        actions = {
            'fsnum': models.UpdateItemAction(
                models.UpdateItemAction.UPDATE_ACTION_PUT,
                models.AttributeValue(
                    models.ATTRIBUTE_TYPE_NUMBER_SET, {42, 43})),
        }

        self.CASANDRA_STORAGE_IMPL.update_item(
            self.context, self.table_name, keys, actions)

        expected = self.expected_data.copy()

        expected['fsnum'] = models.AttributeValue(
            models.ATTRIBUTE_TYPE_NUMBER_SET, {42, 43})

        keys_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, keys_condition)

        self.assertEqual([expected], result.items)

    def test_update_item_put_dynamic_set_blob(self):
        self._create_table(indexed=True)
        self._insert_data()

        keys = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1')
        }

        actions = {
            'fsblb': models.UpdateItemAction(
                models.UpdateItemAction.UPDATE_ACTION_PUT,
                models.AttributeValue(
                    models.ATTRIBUTE_TYPE_BLOB_SET, {'new1', 'new2'})),
        }

        self.CASANDRA_STORAGE_IMPL.update_item(
            self.context, self.table_name, keys, actions)

        expected = self.expected_data.copy()

        expected['fsblb'] = models.AttributeValue(
            models.ATTRIBUTE_TYPE_BLOB_SET, {'new1', 'new2'})

        keys_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, keys_condition)

        self.assertEqual([expected], result.items)

    def test_update_item_delete(self):
        self._create_table(indexed=True)
        self._insert_data()

        keys = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1')
        }

        actions = {
            'str': models.UpdateItemAction(
                models.UpdateItemAction.UPDATE_ACTION_DELETE, None),
            'fstr': models.UpdateItemAction(
                models.UpdateItemAction.UPDATE_ACTION_DELETE, None)
        }

        self.CASANDRA_STORAGE_IMPL.update_item(
            self.context, self.table_name, keys, actions)

        expected = self.expected_data.copy()

        del expected['str']
        del expected['fstr']

        keys_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, keys_condition)

        self.assertEqual([expected], result.items)


class TestCassandraPutItem(TestCassandraBase):
    def test_put_item_str(self):
        self._create_table(indexed=True)

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'str': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, 'str'),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        self.CASANDRA_STORAGE_IMPL.put_item(self.context, put_request)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_number(self):
        self._create_table(indexed=True)

        put = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1'),
            'number': models.AttributeValue.number(42),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        self.CASANDRA_STORAGE_IMPL.put_item(self.context, put_request)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_blob(self):
        self._create_table(indexed=True)

        put = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('1'),
            'blb': models.AttributeValue.blob('blob'),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        self.CASANDRA_STORAGE_IMPL.put_item(self.context, put_request)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_set_str(self):
        self._create_table(indexed=True)

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'set_string': models.AttributeValue(
                models.ATTRIBUTE_TYPE_STRING_SET, {'str1', 'str2'}),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        self.CASANDRA_STORAGE_IMPL.put_item(self.context, put_request)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_set_number(self):
        self._create_table(indexed=True)

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'set_number': models.AttributeValue(
                models.ATTRIBUTE_TYPE_NUMBER_SET, {42, 43}),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        self.CASANDRA_STORAGE_IMPL.put_item(self.context, put_request)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_set_blob(self):
        self._create_table(indexed=True)

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'set_blob': models.AttributeValue(
                models.ATTRIBUTE_TYPE_BLOB_SET, {'blob1', 'blob2'}),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        self.CASANDRA_STORAGE_IMPL.put_item(self.context, put_request)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_dynamic_str(self):
        self._create_table(indexed=True)

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'fstr': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, 'str')
        }

        put_request = models.PutItemRequest(self.table_name, put)

        self.CASANDRA_STORAGE_IMPL.put_item(self.context, put_request)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_dynamic_number(self):
        self._create_table(indexed=True)

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'fnum': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 42),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        self.CASANDRA_STORAGE_IMPL.put_item(self.context, put_request)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_dynamic_blob(self):
        self._create_table(indexed=True)

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'fblb': models.AttributeValue(models.ATTRIBUTE_TYPE_BLOB, 'blob'),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        self.CASANDRA_STORAGE_IMPL.put_item(self.context, put_request)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_dynamic_set_str(self):
        self._create_table(indexed=True)

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'fsstr': models.AttributeValue(
                models.ATTRIBUTE_TYPE_STRING_SET, {'str1', 'str2'}),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        self.CASANDRA_STORAGE_IMPL.put_item(self.context, put_request)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_dynamic_set_number(self):
        self._create_table(indexed=True)

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'fsnum': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER_SET,
                                           {42, 43})
        }

        put_request = models.PutItemRequest(self.table_name, put)

        self.CASANDRA_STORAGE_IMPL.put_item(self.context, put_request)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_dynamic_set_blob(self):
        self._create_table(indexed=True)

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'fsblb': models.AttributeValue(
                models.ATTRIBUTE_TYPE_BLOB_SET, {'blob1', 'blob2'}),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        self.CASANDRA_STORAGE_IMPL.put_item(self.context, put_request)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_expected_str(self):
        self._create_table(indexed=True)
        self._insert_data()

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'str': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, 'str'),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        expected = {
            'str': [
                models.ExpectedCondition.eq(models.AttributeValue.str('str'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.put_item(
            self.context, put_request, expected_condition_map=expected
        )
        self.assertTrue(result)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_expected_number(self):
        self._create_table(indexed=True)
        self._insert_data()

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'numbr': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 42),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        expected = {
            'numbr': [
                models.ExpectedCondition.eq(models.AttributeValue.number(1))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.put_item(
            self.context, put_request, expected_condition_map=expected
        )
        self.assertTrue(result)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_expected_blob(self):
        self._create_table(indexed=True)
        self._insert_data()

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'blb': models.AttributeValue(models.ATTRIBUTE_TYPE_BLOB, 'blob'),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        expected = {
            'blb': [
                models.ExpectedCondition.eq(models.AttributeValue.blob('blob'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.put_item(
            self.context, put_request, expected_condition_map=expected
        )
        self.assertTrue(result)

        key_condition = {
            'id': [
                models.Condition.eq(models.AttributeValue.number(1))
            ],
            'range': [
                models.Condition.eq(models.AttributeValue.str('1'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_expected_set_str(self):
        self._create_table(indexed=True)
        self._insert_data()

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'set_string': models.AttributeValue(
                models.ATTRIBUTE_TYPE_STRING_SET, {'str1', 'str2'}),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        expected = {
            'set_string': [
                models.ExpectedCondition.eq(models.AttributeValue.str_set(
                    {'a', 'b', 'c'}))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.put_item(
            self.context, put_request, expected_condition_map=expected
        )
        self.assertTrue(result)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_expected_set_number(self):
        self._create_table(indexed=True)
        self._insert_data()

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'set_number': models.AttributeValue(
                models.ATTRIBUTE_TYPE_NUMBER_SET, {42, 43}),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        expected = {
            'set_number': [
                models.ExpectedCondition.eq(models.AttributeValue.number_set(
                    {1, 2, 3}))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.put_item(
            self.context, put_request, expected_condition_map=expected
        )
        self.assertTrue(result)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_expected_set_blob(self):
        self._create_table(indexed=True)
        self._insert_data()

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'set_blob': models.AttributeValue(
                models.ATTRIBUTE_TYPE_BLOB_SET, {'blob1', 'blob2'}),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        expected = {
            'set_blob': [
                models.ExpectedCondition.eq(models.AttributeValue.blob_set(
                    {'blob1', 'blob2'}))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.put_item(
            self.context, put_request, expected_condition_map=expected
        )
        self.assertTrue(result)

        key_condition = {
            'id': [
                models.Condition.eq(models.AttributeValue.number(1))
            ],
            'range': [
                models.Condition.eq(models.AttributeValue.str('1'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_expected_dynamic_str(self):
        self._create_table(indexed=True)
        self._insert_data()

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'fstr': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, 'str'),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        expected = {
            'fstr': [
                models.ExpectedCondition.eq(models.AttributeValue.str('fstr'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.put_item(
            self.context, put_request, expected_condition_map=expected
        )
        self.assertTrue(result)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_expected_dynamic_number(self):
        self._create_table(indexed=True)
        self._insert_data()

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'fnum': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 42),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        expected = {
            'fnum': [
                models.ExpectedCondition.eq(models.AttributeValue.number(1))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.put_item(
            self.context, put_request, expected_condition_map=expected
        )
        self.assertTrue(result)

        key_condition = {
            'id': [
                models.Condition.eq(models.AttributeValue.number(1))
            ],
            'range': [
                models.Condition.eq(models.AttributeValue.str('1'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_expected_dynamic_blob(self):
        self._create_table(indexed=True)
        self._insert_data()

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'fblb': models.AttributeValue(models.ATTRIBUTE_TYPE_BLOB, 'blob'),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        expected = {
            'fblb': [
                models.ExpectedCondition.eq(
                    models.AttributeValue.blob('fblob'))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.put_item(
            self.context, put_request, expected_condition_map=expected
        )
        self.assertTrue(result)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_expected_dynamic_set_str(self):
        self._create_table(indexed=True)
        self._insert_data()

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'fsstr': models.AttributeValue(
                models.ATTRIBUTE_TYPE_STRING_SET, {'str1', 'str2'}),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        expected = {
            'fsstr': [
                models.ExpectedCondition.eq(
                    models.AttributeValue.str_set({'fa', 'fb', 'fc'}))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.put_item(
            self.context, put_request, expected_condition_map=expected
        )
        self.assertTrue(result)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_expected_dynamic_set_number(self):
        self._create_table(indexed=True)
        self._insert_data()

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'fsnum': models.AttributeValue(
                models.ATTRIBUTE_TYPE_NUMBER_SET, {42, 43}),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        expected = {
            'fsnum': [
                models.ExpectedCondition.eq(
                    models.AttributeValue.number_set({1, 2, 3}))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.put_item(
            self.context, put_request, expected_condition_map=expected
        )
        self.assertTrue(result)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)

    def test_put_item_expected_dynamic_set_blob(self):
        self._create_table(indexed=True)
        self._insert_data()

        put = {
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'fsblb': models.AttributeValue(
                models.ATTRIBUTE_TYPE_BLOB_SET, {'blob1', 'blob2'}),
        }

        put_request = models.PutItemRequest(self.table_name, put)

        expected = {
            'fsblob': [
                models.ExpectedCondition.eq(
                    models.AttributeValue.blob_set({'fblob1', 'fblob2'}))
            ]
        }

        result = self.CASANDRA_STORAGE_IMPL.put_item(
            self.context, put_request, expected_condition_map=expected
        )
        self.assertTrue(result)

        key_condition = {
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.select_item(
            self.context, self.table_name, key_condition)

        self.assertEqual([put], result.items)


class TestCassandraScan(TestCassandraBase):

    def test_scan(self):
        self._create_table(indexed=True)
        self._insert_data()

        result = self.CASANDRA_STORAGE_IMPL.scan(
            self.context, self.table_name, {})

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_scan_not_equal(self):
        self._create_table(indexed=True)
        self._insert_data()
        self._insert_data(range_value='2')

        condition = {
            'range': [models.ScanCondition.neq(models.AttributeValue.str('2'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.scan(
            self.context, self.table_name, condition)

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_scan_contains(self):
        self._create_table(indexed=True)
        self._insert_data()
        self._insert_data(range_value='121')

        condition = {
            'range': [models.ScanCondition.contains(
                models.AttributeValue.str('2'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.scan(
            self.context, self.table_name, condition)

        self.assertEqual(1, result.count)
        self.assertEqual('121', result.items[0]['range'].value)

    def test_scan_not_contains(self):
        self._create_table(indexed=True)
        self._insert_data()
        self._insert_data(range_value='22')

        condition = {
            'range': [models.ScanCondition.not_contains(
                models.AttributeValue.str('2'))]
        }

        result = self.CASANDRA_STORAGE_IMPL.scan(
            self.context, self.table_name, condition)

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_scan_contains_for_set(self):
        self._create_table(indexed=True)
        self._insert_data()

        condition = {
            'set_number': [models.ScanCondition.contains(
                models.AttributeValue.number(2))]
        }

        result = self.CASANDRA_STORAGE_IMPL.scan(
            self.context, self.table_name, condition)

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_scan_contains_for_set2(self):
        self._create_table(indexed=True)
        self._insert_data()

        condition = {
            'set_number': [models.ScanCondition.contains(
                models.AttributeValue.number(4))]
        }

        result = self.CASANDRA_STORAGE_IMPL.scan(
            self.context, self.table_name, condition)

        self.assertEqual(0, result.count)

    def test_scan_not_contains_for_set(self):
        self._create_table(indexed=True)
        self._insert_data()

        condition = {
            'set_number': [models.ScanCondition.not_contains(
                models.AttributeValue.number(4))]
        }

        result = self.CASANDRA_STORAGE_IMPL.scan(
            self.context, self.table_name, condition)

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_scan_not_contains_for_set2(self):
        self._create_table(indexed=True)
        self._insert_data()

        condition = {
            'set_number': [models.ScanCondition.not_contains(
                models.AttributeValue.number(2))]
        }

        result = self.CASANDRA_STORAGE_IMPL.scan(
            self.context, self.table_name, condition)

        self.assertEqual(0, result.count)

    def test_scan_in(self):
        self._create_table(indexed=True)
        self._insert_data()

        condition = {
            'range': [models.ScanCondition.in_set({
                models.AttributeValue.str('1'),
                models.AttributeValue.str('2')
            })]
        }

        result = self.CASANDRA_STORAGE_IMPL.scan(
            self.context, self.table_name, condition)

        self.assertEqual(1, result.count)
        self._validate_data(result.items[0])

    def test_scan_in_negative(self):
        self._create_table(indexed=True)
        self._insert_data()

        condition = {
            'range': [models.ScanCondition.in_set({
                models.AttributeValue.str('2'),
                models.AttributeValue.str('3')
            })]
        }

        result = self.CASANDRA_STORAGE_IMPL.scan(
            self.context, self.table_name, condition)

        self.assertEqual(0, result.count)

    def test_paging(self):
        self._create_table(indexed=True)
        self._insert_data()
        self._insert_data(range_value='2')
        self._insert_data(range_value='3')
        self._insert_data(id_value=2, range_value='1',)
        self._insert_data(id_value=2, range_value='2')

        last_evaluated_key = {
            'id': models.AttributeValue.number(1),
            'range': models.AttributeValue.str('2')
        }

        result = self.CASANDRA_STORAGE_IMPL.scan(
            self.context, self.table_name, {},
            exclusive_start_key=last_evaluated_key,
            limit=2)

        last_evaluated_key2 = {
            'id': models.AttributeValue.number(2),
            'range': models.AttributeValue.str('1')
        }

        result2 = self.CASANDRA_STORAGE_IMPL.scan(
            self.context, self.table_name, {},
            exclusive_start_key=last_evaluated_key2,
            limit=2)

        if result.count == 2:
            self.assertTrue(result.items[0]['id'].value == 1 and
                            result.items[0]['range'].value == '3')

            self.assertTrue(result.items[1]['id'].value == 2 and
                            result.items[1]['range'].value == '1')

        elif result2.count == 2:
            self.assertTrue(result2.items[0]['id'].value == 2 and
                            result2.items[0]['range'].value == '2')

            self.assertTrue(result2.items[1]['id'].value == 1 and
                            result2.items[1]['range'].value == '1')

        else:
            self.fail()


class TestCassandraBatch(TestCassandraBase):
    def test_batch_write_item(self):
        self._create_table(indexed=True)
        put_items = [{
            'id': models.AttributeValue(models.ATTRIBUTE_TYPE_NUMBER, 1),
            'range': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, '1'),
            'str': models.AttributeValue(models.ATTRIBUTE_TYPE_STRING, 'str'),
        }, {
            'id': models.AttributeValue.number(2),
            'range': models.AttributeValue.str('2'),
            'number': models.AttributeValue.number(42),
        }, {
            'id': models.AttributeValue.number(3),
            'range': models.AttributeValue.str('3'),
            'blb': models.AttributeValue.blob('blob'),
        }]

        put_requests = [models.PutItemRequest(self.table_name, i)
                        for i in put_items]

        self.CASANDRA_STORAGE_IMPL.execute_write_batch(self.context,
                                                       put_requests)

        key_conditions = [{
            'id': [models.Condition.eq(models.AttributeValue.number(1))],
            'range': [models.Condition.eq(models.AttributeValue.str('1'))]
        }, {
            'id': [models.Condition.eq(models.AttributeValue.number(2))],
            'range': [models.Condition.eq(models.AttributeValue.str('2'))]
        }, {
            'id': [models.Condition.eq(models.AttributeValue.number(3))],
            'range': [models.Condition.eq(models.AttributeValue.str('3'))]
        }]

        for key, item in zip(key_conditions, put_items):
            result = self.CASANDRA_STORAGE_IMPL.select_item(
                self.context, self.table_name, key)

            self.assertEqual([item], result.items)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import httplib
import json

import mock

from magnetodb.tests.unittests.api.openstack.v1 import test_base_testcase


class DeleteItemTestCase(test_base_testcase.APITestCase):
    """The test for delete_item method of openstack v1 ReST API."""

    @mock.patch('magnetodb.storage.delete_item')
    def test_delete_item(self, mock_delete_item):
        headers = {'Content-Type': 'application/json',
                   'Accept': 'application/json'}

        conn = httplib.HTTPConnection('localhost:8080')
        url = '/v1/fake_project_id/data/tables/the_table/delete_item'

        # basic test case: delete item by key
        body = """
            {
                "key": {
                    "ForumName": {
                        "S": "MagnetoDB"
                    },
                    "Subject": {
                        "S": "How do I delete an item?"
                    }
                }
            }
        """

        conn.request("POST", url, headers=headers, body=body)
        response = conn.getresponse()
        self.assertTrue(mock_delete_item.called)
        self.assertEqual(200, response.status)

        json_response = response.read()
        response_payload = json.loads(json_response)
        self.assertEqual({}, response_payload)

    @mock.patch('magnetodb.storage.delete_item')
    def test_delete_item_with_expected_conditions(self, mock_delete_item):
        headers = {'Content-Type': 'application/json',
                   'Accept': 'application/json'}

        conn = httplib.HTTPConnection('localhost:8080')
        url = '/v1/fake_project_id/data/tables/the_table/delete_item'

        # conditional delete item: in addition to key,
        # expected conditions and/or return values can be specified
        # expected conditions represents an attribute name to check,
        # along with value comparison
        # or value evaluation before attempting the conditional delete

        body = """
            {
                "key": {
                    "ForumName": {
                        "S": "MagnetoDB"
                    },
                    "Subject": {
                        "S": "How do I delete an item?"
                    }
                },
                "expected": {
                    "Subject": {
                        "value": {
                            "S": "How do I delete an item?"
                        }
                    },
                    "Replies": {
                        "exists": false
                    }
                },
                "returnValues": ""
            }
        """

        conn.request("POST", url, headers=headers, body=body)
        response = conn.getresponse()
        self.assertTrue(mock_delete_item.called)
        self.assertEqual(200, response.status)

        json_response = response.read()
        response_payload = json.loads(json_response)
        self.assertEqual({}, response_payload)

    def test_delete_item_malformed(self):
        headers = {'Content-Type': 'application/json',
                   'Accept': 'application/json'}

        conn = httplib.HTTPConnection('localhost:8080')
        url = '/v1/fake_project_id/data/tables/the_table/delete_item'
        body = '{"foo": "bar"}'
        conn.request("POST", url, headers=headers, body=body)
        response = conn.getresponse()

        json_response = response.read()
        self.assertEqual(400, response.status)
        response_payload = json.loads(json_response)

        expected_message = "'key' is a required property"
        expected_type = 'ValidationError'

        self.assertEqual(expected_message,
                         response_payload['error']['message'])
        self.assertEqual(expected_type,
                         response_payload['error']['type'])



# Copyright 2013 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import jsonschema

from magnetodb.common import exception
from magnetodb.openstack.common.log import logging

LOG = logging.getLogger(__name__)


class DynamoDBAction():
    schema = {}

    def __init__(self, context, action_params):
        self.action_params = action_params
        self.context = context

    @classmethod
    def format_validation_msg(self, errors):
        # format path like object['field1'][i]['subfield2']
        messages = []
        for error in errors:
            path = list(error.path)
            f_path = "%s%s" % (path[0],
                               ''.join(['[%r]' % i for i in path[1:]]))
            messages.append("%s %s" % (f_path, error.message))
            for suberror in sorted(error.context, key=lambda e: e.schema_path):
                messages.append(suberror.message)
        error_msg = "; ".join(messages)
        return "Validation error: %s" % error_msg

    @classmethod
    def validate_params(cls, params):
            assert isinstance(params, dict)

            validator = jsonschema.Draft4Validator(cls.schema)
            if not validator.is_valid(params):
                errors = sorted(validator.iter_errors(params),
                                key=lambda e: e.path)
                error_msg = cls.format_validation_msg(errors)
                LOG.info(error_msg)
                raise exception.ValidationException(error_msg)

    @classmethod
    def perform(cls, context, action_params):
        cls.validate_params(action_params)

        return cls(context, action_params)()



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import httplib
import json

import mock
from magnetodb.storage import models
from magnetodb.tests.unittests.api.openstack.v1 import test_base_testcase


class QueryTest(test_base_testcase.APITestCase):
    """The test for v1 ReST API ScanController."""

    @mock.patch('magnetodb.storage.select_item')
    def test_query(self, mock_query):

        items = [
            {
                'ForumName': models.AttributeValue.str('Testing OS API'),
                'LastPostDateTime': models.AttributeValue.str('3/18/14'),
                'Posts': models.AttributeValue.str_set(['Opening post'])
            },
            {
                'ForumName': models.AttributeValue.str('Testing OS API'),
                'LastPostDateTime': models.AttributeValue.str('3/19/14'),
                'Posts': models.AttributeValue.str_set(['Hi', 'Hello'])
            },
        ]

        last_evaluated_key = {
            'ForumName': models.AttributeValue.str('Testing OS API'),
            'LastPostDateTime': models.AttributeValue.str('3/19/14'),
        }

        mock_query.return_value = models.SelectResult(
            items=items,
            last_evaluated_key=last_evaluated_key,
            count=2)

        headers = {'Content-Type': 'application/json',
                   'Accept': 'application/json'}

        conn = httplib.HTTPConnection('localhost:8080')
        url = '/v1/fake_project_id/data/tables/Threads/query'
        body = """
            {
               "attributes_to_get": [
                   "ForumName", "LastPostDateTime", "Posts"
               ],
               "exclusive_start_key":
                   {
                       "ForumName" :
                           {
                               "S": "Testing OS API"
                           },
                       "LastPostDayTime" :
                           {
                               "S": "3/1/14"
                           }
                   },
               "index_name": "LastPostIndex",
               "limit": 2,
               "consistent_read": true,
               "scan_index_forward": true,
               "key_conditions":
                   {
                        "ForumName" :
                           {
                               "attribute_value_list": [
                                   {
                                       "S": "Testing OS API"
                                   }
                               ],
                               "comparison_operator": "EQ"
                           },
                       "LastPostDateTime" :
                           {
                               "attribute_value_list": [
                                   {
                                       "S": "3/10/14"
                                   }
                               ],
                               "comparison_operator": "GT"
                           }
                   },
               "select": "SPECIFIC_ATTRIBUTES"
            }
        """

        expected_response = {
            "count": 2,
            "items": [
                {
                    'ForumName': {'S': 'Testing OS API'},
                    'LastPostDateTime': {'S': '3/18/14'},
                    'Posts': {'SS': ['Opening post']}
                },
                {
                    'ForumName': {'S': 'Testing OS API'},
                    'LastPostDateTime': {'S': '3/19/14'},
                    'Posts': {'SS': ['Hi', 'Hello']}
                },
            ],
            "last_evaluated_key": {
                'ForumName': {'S': 'Testing OS API'},
                'LastPostDateTime': {'S': '3/19/14'},
            }
        }

        conn.request("POST", url, headers=headers, body=body)

        response = conn.getresponse()

        self.assertTrue(mock_query.called)

        json_response = response.read()
        response_payload = json.loads(json_response)

        self.assertEqual(expected_response, response_payload)

    @mock.patch('magnetodb.storage.select_item')
    def test_query_count(self, mock_query):
        mock_query.return_value = models.SelectResult(count=100500)

        headers = {'Content-Type': 'application/json',
                   'Accept': 'application/json'}

        conn = httplib.HTTPConnection('localhost:8080')
        url = '/v1/fake_project_id/data/tables/Threads/query'
        body = """
            {
                "key_conditions":
                   {
                        "ForumName" :
                           {
                               "attribute_value_list": [
                                   {
                                       "S": "Testing OS API"
                                   }
                               ],
                               "comparison_operator": "EQ"
                           },
                       "LastPostDateTime" :
                           {
                               "attribute_value_list": [
                                   {
                                       "S": "3/10/14"
                                   }
                               ],
                               "comparison_operator": "GT"
                           }
                   },
               "select": "COUNT"
            }
        """

        expected_response = {
            "count": 100500,
        }

        conn.request("POST", url, headers=headers, body=body)

        response = conn.getresponse()

        self.assertTrue(mock_query.called)

        json_response = response.read()
        response_payload = json.loads(json_response)

        self.assertEqual(expected_response, response_payload)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from magnetodb import storage
from magnetodb.openstack.common.log import logging

from magnetodb.api.openstack.v1 import parser
from magnetodb.api.openstack.v1 import utils


LOG = logging.getLogger(__name__)


class DeleteTableController(object):
    def delete_table(self, req, project_id, table_name):
        utils.check_project_id(req.context, project_id)
        req.context.tenant = project_id

        table_schema = storage.delete_table(req.context, table_name)

        url = req.path_url
        bookmark = req.path_url

        return {
            parser.Props.TABLE_DESCRIPTION: {
                parser.Props.ATTRIBUTE_DEFINITIONS: (
                    parser.Parser.format_attribute_definitions(
                        table_schema.schema.attribute_type_map
                    )
                ),
                parser.Props.CREATION_DATE_TIME: 0,
                parser.Props.ITEM_COUNT: 0,
                parser.Props.KEY_SCHEMA: (
                    parser.Parser.format_key_schema(
                        table_schema.schema.key_attributes
                    )
                ),
                parser.Props.LOCAL_SECONDARY_INDEXES: (
                    parser.Parser.format_local_secondary_indexes(
                        table_schema.schema.key_attributes[0],
                        table_schema.schema.index_def_map
                    )
                ),
                parser.Props.TABLE_NAME: table_name,

                parser.Props.TABLE_STATUS: (
                    parser.Parser.format_table_status(table_schema.status)),
                parser.Props.TABLE_SIZE_BYTES: 0,

                parser.Props.LINKS: [
                    {
                        parser.Props.HREF: url,
                        parser.Props.REL: parser.Values.SELF
                    },
                    {
                        parser.Props.HREF: bookmark,
                        parser.Props.REL: parser.Values.BOOKMARK
                    }
                ]
            }
        }



# Copyright 2013 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo.config import cfg

from magnetodb.common import PROJECT_NAME

common_opts = [
    cfg.StrOpt('storage_manager_config', default="{}")
]

CONF = cfg.CONF
CONF.register_opts(common_opts)


def parse_args(prog=None, args=None, default_config_files=None):
    cfg.CONF(args=args,
             project=PROJECT_NAME,
             prog=prog,
             default_config_files=default_config_files)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.api.keyvalue.boto_base.base import MagnetoDBTestCase
from tempest.common.utils.data_utils import rand_name
from tempest.test import attr


class MagnetoDBScanTest(MagnetoDBTestCase):

    def setUp(self):
        super(MagnetoDBScanTest, self).setUp()
        self.tname = rand_name().replace('-', '')
        self.client.create_table(self.smoke_attrs,
                                 self.tname,
                                 self.smoke_schema,
                                 self.smoke_throughput)
        self.wait_for_table_active(self.tname)
        self.rck = self.addResourceCleanUp(self.client.delete_table,
                                           self.tname)

    def tearDown(self):
        super(MagnetoDBScanTest, self).tearDown()
        self.client.delete_table(self.tname)
        self.assertTrue(self.wait_for_table_deleted(self.tname))
        self.cancelResourceCleanUp(self.rck)

    def _verify_scan_response(self, response,
                              attributes_to_get=None, limit=None,
                              select=None, scan_filter=None,
                              exclusive_start_key=None,
                              return_consumed_capacity=None,
                              total_segments=None, segment=None):
        """
        The method implements common verifications of response format.

        More specific verifications should be implemented in the test-cases
        explicitly.
        """
        if return_consumed_capacity is None:
            self.assertNotIn('ConsumedCapacity', response)
        self.assertIn('ScannedCount', response)
        self.assertIn('Count', response)
        self.assertEqual(select == 'COUNT', 'Items' not in response)
        if select != 'COUNT':
            self.assertEqual(response['Count'], len(response['Items']))
        if limit is None:
            self.assertNotIn('LastEvaluatedKey', response)

    @attr(type='smoke')
    def test_scan(self):
        item = self.put_smoke_item(self.tname, 'forum1', 'subject2')
        resp = self.client.scan(table_name=self.tname)
        self._verify_scan_response(resp)
        self.assertEqual(1, resp['Count'])
        self.assertEqual(item, resp['Items'][0])

    def test_scan_limit_scanfilter(self):
        """
        Verifies the scan operation processes all existing items,
        filters appropriate ones and stops scanning if the limit reached.
        """
        exp_scanned_count = 2 * 10
        exp_count = 10
        limit = 5
        self.populate_smoke_table(self.tname, 2, exp_count)

        scanfilter = {
            'message': {
                'ComparisonOperator': 'NOT_NULL'
            }
        }
        resp1 = self.client.scan(self.tname, scan_filter=scanfilter,
                                 limit=limit)

        self._verify_scan_response(response=resp1, limit=limit,
                                   scan_filter=scanfilter)
        self.assertEqual(limit, resp1['ScannedCount'])
        count1 = resp1['Count']
        last_key = resp1['LastEvaluatedKey']

        resp2 = self.client.scan(self.tname, scan_filter=scanfilter,
                                 exclusive_start_key=last_key)
        self.assertEqual(exp_scanned_count - limit, resp2['ScannedCount'])
        count2 = resp2['Count']
        self.assertEqual(exp_scanned_count, count1 + count2)
        for item in resp1['Items']:
            self.assertNotIn(item, resp2['Items'])

    def test_scan_attr_to_get_select(self):
        """
        Test the cases:
          - AttributesToGet contains non-key attr not used in filter;
          - AttributesToGet contains nonexistent attr;
          - Select=SPECIFIC_ATTRIBUTES and AttributesToGet specified;
          - Select=ALL_ATTRIBUTES;
          - Select=COUNT.
        """
        item1 = self.put_smoke_item(self.tname, 'forum', 'subject1',
                                    'filtered', 'John')
        self.put_smoke_item(self.tname, 'forum', 'subject2', 'skipped', 'Alex')

        attrs_to_get = ['last_posted_by', 'nonexistent_attr']
        scanfilter = {
            'message': {
                'AttributeValueList': [{'S': 'filter'}],
                'ComparisonOperator': 'CONTAINS'
            }
        }

        resp1 = self.client.scan(self.tname, scan_filter=scanfilter,
                                 attributes_to_get=attrs_to_get)
        self.assertEqual(1, resp1['Count'])

        exp_items = [{'last_posted_by': item1['last_posted_by']}]
        self.assertEqual(exp_items, resp1['Items'])

        # If Select=SPECIFIC_ATTRIBUTES and AttributesToGet are specified,
        # result should be equivalent to specifying AttributesToGet without
        # specifying any value for Select.
        resp2 = self.client.scan(self.tname, scan_filter=scanfilter,
                                 attributes_to_get=attrs_to_get,
                                 select='SPECIFIC_ATTRIBUTES')
        self.assertEqual(resp1, resp2)

        # verify results if Select set to get all attributes
        resp3 = self.client.scan(self.tname, scan_filter=scanfilter,
                                 select='ALL_ATTRIBUTES')
        self.assertEqual([item1], resp3['Items'])

        resp4 = self.client.scan(self.tname, select='COUNT')
        self._verify_scan_response(resp4, select='COUNT')
        self.assertEqual(2, resp4['Count'])



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the 'License'); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an 'AS IS' BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random
import string

from tempest import exceptions
from tempest.api.keyvalue.rest_base.base import MagnetoDBTestCase
from tempest.test import attr


class MagnetoDBGetItemTest(MagnetoDBTestCase):
    @classmethod
    def setUpClass(cls):
        super(MagnetoDBGetItemTest, cls).setUpClass()

    def random_name(self, length):
        return ''.join(random.choice(string.lowercase + string.digits)
                       for i in range(length))

    @attr(type='GI-12')
    def test_get_item_long_table_name(self):
        table_name = self.random_name(255)
        self._create_test_table(self.smoke_attrs + self.index_attrs,
                                table_name,
                                self.smoke_schema,
                                wait_for_active=True)
        item = self.build_smoke_item('forum1', 'subject2',
                                     last_posted_by='John')
        key = {self.hashkey: item[self.hashkey],
               self.rangekey: item[self.rangekey]}

        self.put_smoke_item(table_name, 'forum1', 'subject2')

        attributes_to_get = ['last_posted_by']
        get_resp = self.client.get_item(table_name,
                                        key,
                                        attributes_to_get,
                                        True)
        self.assertEqual(get_resp[1]['item']['last_posted_by'], {'S': 'John'})

    @attr(type=['GI-13', 'negative'])
    def test_get_item_non_existent_table_name(self):
        item = self.build_smoke_item('forum1', 'subject2',
                                     last_posted_by='John')
        key = {self.hashkey: item[self.hashkey],
               self.rangekey: item[self.rangekey]}

        attributes_to_get = ['last_posted_by']
        with self.assertRaises(exceptions.NotFound) as raises_cm:
            self.client.get_item("nonexistenttable",
                                 key,
                                 attributes_to_get,
                                 True)
        exception = raises_cm.exception
        self.assertIn("Not Found",
                      exception._error_string)
        self.assertIn("The resource could not be found.",
                      exception._error_string)
        self.assertIn("Table 'nonexistenttable' does not exists",
                      exception._error_string)

    # TODO(ValidationException or simple NotFound)
    @attr(type=['GI-15', 'negative'])
    def test_get_item_two_symbol_table_name(self):
        item = self.build_smoke_item('forum1', 'subject2',
                                     last_posted_by='John')
        key = {self.hashkey: item[self.hashkey],
               self.rangekey: item[self.rangekey]}

        attributes_to_get = ['last_posted_by']
        with self.assertRaises(exceptions.NotFound):
            self.client.get_item(self.random_name(2),
                                 key,
                                 attributes_to_get,
                                 True)

    # TODO(ValidationException or simple NotFound)
    @attr(type=['GI-16', 'negative'])
    def test_get_item_more_than_255_symbol_table_name(self):
        item = self.build_smoke_item('forum1', 'subject2',
                                     last_posted_by='John')
        key = {self.hashkey: item[self.hashkey],
               self.rangekey: item[self.rangekey]}

        attributes_to_get = ['last_posted_by']
        with self.assertRaises(exceptions.NotFound):
            self.client.get_item(self.random_name(260),
                                 key,
                                 attributes_to_get,
                                 True)

    @attr(type=['GI-51', 'negative'])
    def test_get_item_resource_not_found_exception(self):
        item = self.build_smoke_item('forum1', 'subject2',
                                     last_posted_by='John')
        key = {self.hashkey: item[self.hashkey],
               self.rangekey: item[self.rangekey]}

        attributes_to_get = ['last_posted_by']
        with self.assertRaises(exceptions.NotFound) as raises_cm:
            self.client.get_item("nonexistenttable",
                                 key,
                                 attributes_to_get,
                                 True)
        exception = raises_cm.exception
        self.assertIn("ResourceNotFoundException", exception._error_string)



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.api.keyvalue.rest_base.base import MagnetoDBTestCase
from tempest.common.utils.data_utils import rand_name
from tempest.test import attr


class MagnetoDBBatchWriteTest(MagnetoDBTestCase):

    def setUp(self):
        super(MagnetoDBBatchWriteTest, self).setUp()
        self.tname = rand_name().replace('-', '')

    @attr(type=['BWI-1'])
    def test_batch_write_mandatory(self):
        self._create_test_table(self.build_x_attrs('S'),
                                self.tname,
                                self.smoke_schema,
                                wait_for_active=True)
        item = self.build_x_item('S', 'forum1', 'subject2',
                                 ('message', 'S', 'message text'))
        request_body = {'request_items': {self.tname: [{'put_request':
                                                        {'item': item}}]}}
        headers, body = self.client.batch_write_item(request_body)
        self.assertIn('unprocessed_items', body)
        self.assertEqual(body['unprocessed_items'], {})
        key_conditions = {
            self.hashkey: {
                'attribute_value_list': [{'S': 'forum1'}],
                'comparison_operator': 'EQ'
            }
        }
        headers, body = self.client.query(self.tname,
                                          key_conditions=key_conditions,
                                          consistent_read=True)
        self.assertEqual(item, body['items'][0])

    @attr(type=['BWI-2'])
    def test_batch_write_all_params(self):
        self._create_test_table(self.build_x_attrs('S'),
                                self.tname,
                                self.smoke_schema,
                                wait_for_active=True)
        item = self.build_x_item('S', 'forum1', 'subject2',
                                 ('message', 'S', 'message text'))
        request_body = {'request_items': {self.tname: [{'put_request':
                                                        {'item': item}}]}}
        headers, body = self.client.batch_write_item(request_body)
        self.assertIn('unprocessed_items', body)
        self.assertEqual(body['unprocessed_items'], {})
        key_conditions = {
            self.hashkey: {
                'attribute_value_list': [{'S': 'forum1'}],
                'comparison_operator': 'EQ'
            }
        }
        headers, body = self.client.query(self.tname,
                                          key_conditions=key_conditions,
                                          consistent_read=True)
        self.assertEqual(item, body['items'][0])

    @attr(type=['BWI-10'])
    def test_batch_write_put_n(self):
        self._create_test_table(self.build_x_attrs('N'),
                                self.tname,
                                self.smoke_schema,
                                wait_for_active=True)
        item = self.build_x_item('N', '1000', '2000', ('message', 'N', '1001'))
        request_body = {'request_items': {self.tname: [{'put_request':
                                                        {'item': item}}]}}
        headers, body = self.client.batch_write_item(request_body)
        self.assertIn('unprocessed_items', body)
        self.assertEqual(body['unprocessed_items'], {})
        key_conditions = {
            self.hashkey: {
                'attribute_value_list': [{'N': '1000'}],
                'comparison_operator': 'EQ'
            }
        }
        headers, body = self.client.query(self.tname,
                                          key_conditions=key_conditions,
                                          consistent_read=True)
        self.assertEqual(item, body['items'][0])

    @attr(type=['BWI-13'])
    def test_batch_write_put_n_ns(self):
        self._create_test_table(self.build_x_attrs('N'),
                                self.tname,
                                self.smoke_schema,
                                wait_for_active=True)
        item = self.build_x_item('N', '1000', '2000',
                                 ('message', 'NS', ['1001', '1002']))
        request_body = {'request_items': {self.tname: [{'put_request':
                                                        {'item': item}}]}}
        headers, body = self.client.batch_write_item(request_body)
        self.assertIn('unprocessed_items', body)
        self.assertEqual(body['unprocessed_items'], {})
        key_conditions = {
            self.hashkey: {
                'attribute_value_list': [{'N': '1000'}],
                'comparison_operator': 'EQ'
            }
        }
        headers, body = self.client.query(self.tname,
                                          key_conditions=key_conditions,
                                          consistent_read=True)
        self.assertEqual(item, body['items'][0])

    @attr(type=['BWI-36'])
    def test_batch_write_delete_s(self):
        self._create_test_table(self.smoke_attrs,
                                self.tname,
                                self.smoke_schema,
                                wait_for_active=True)
        self.put_smoke_item(self.tname, 'forum1', 'subject2',
                            'message text', 'John', '10')
        key_conditions = {
            self.hashkey: {
                'attribute_value_list': [{'S': 'forum1'}],
                'comparison_operator': 'EQ'
            },
            self.rangekey: {
                'attribute_value_list': [{'S': 'subject'}],
                'comparison_operator': 'BEGINS_WITH'
            }
        }
        headers, body = self.client.query(self.tname,
                                          key_conditions=key_conditions,
                                          consistent_read=True)
        self.assertEqual(body['count'], 1)

        key = {self.hashkey: {'S': 'forum1'}, self.rangekey: {'S': 'subject2'}}
        request_body = {'request_items': {self.tname: [{'delete_request':
                                                        {'key': key}}]}}
        headers, body = self.client.batch_write_item(request_body)
        self.assertIn('unprocessed_items', body)
        self.assertEqual(body['unprocessed_items'], {})
        headers, body = self.client.query(self.tname,
                                          key_conditions=key_conditions,
                                          consistent_read=True)
        self.assertEqual(0, body['count'])

    @attr(type=['BWI-53'])
    def test_batch_write_delete(self):
        self._create_test_table(self.smoke_attrs,
                                self.tname,
                                self.smoke_schema,
                                wait_for_active=True)
        self.put_smoke_item(self.tname, 'forum1', 'subject2',
                            'message text', 'John', '10')
        key_conditions = {
            self.hashkey: {
                'attribute_value_list': [{'S': 'forum1'}],
                'comparison_operator': 'EQ'
            },
            self.rangekey: {
                'attribute_value_list': [{'S': 'subject'}],
                'comparison_operator': 'BEGINS_WITH'
            }
        }
        headers, body = self.client.query(self.tname,
                                          key_conditions=key_conditions,
                                          consistent_read=True)
        self.assertEqual(body['count'], 1)

        key = {self.hashkey: {'S': 'forum1'}, self.rangekey: {'S': 'subject2'}}
        request_body = {'request_items': {self.tname: [{'delete_request':
                                                        {'key': key}}]}}
        headers, body = self.client.batch_write_item(request_body)
        self.assertIn('unprocessed_items', body)
        self.assertEqual(body['unprocessed_items'], {})
        headers, body = self.client.query(self.tname,
                                          key_conditions=key_conditions,
                                          consistent_read=True)
        self.assertEqual(0, body['count'])

    @attr(type=['BWI-40'])
    def test_batch_write_put_delete(self):
        self._create_test_table(self.smoke_attrs,
                                self.tname,
                                self.smoke_schema,
                                wait_for_active=True)
        self.put_smoke_item(self.tname, 'forum1', 'subject2',
                            'message text', 'John', '10')
        key_conditions = {
            self.hashkey: {
                'attribute_value_list': [{'S': 'forum1'}],
                'comparison_operator': 'EQ'
            },
            self.rangekey: {
                'attribute_value_list': [{'S': 'subject'}],
                'comparison_operator': 'BEGINS_WITH'
            }
        }
        headers, body = self.client.query(self.tname,
                                          key_conditions=key_conditions,
                                          consistent_read=True)
        self.assertEqual(body['count'], 1)

        item = self.build_smoke_item('forum1', 'subject3',
                                     'message text', 'John', '10')

        key = {self.hashkey: {'S': 'forum1'}, self.rangekey: {'S': 'subject2'}}
        request_body = {'request_items': {self.tname: [{'delete_request':
                                                        {'key': key}},
                                                       {'put_request':
                                                        {'item': item}}]}}
        headers, body = self.client.batch_write_item(request_body)
        self.assertIn('unprocessed_items', body)
        self.assertEqual(body['unprocessed_items'], {})
        headers, body = self.client.query(self.tname,
                                          key_conditions=key_conditions,
                                          consistent_read=True)
        self.assertEqual(1, body['count'])
        self.assertIn(item, body['items'])

    @attr(type=['BWI-51'])
    def test_batch_write_put_25_tables(self):
        tnames = [rand_name().replace('-', '') for i in range(0, 25)]
        for tname in tnames:
            self._create_test_table(self.smoke_attrs,
                                    tname,
                                    self.smoke_schema,
                                    wait_for_active=True)
        item = self.build_smoke_item('forum1', 'subject3',
                                     'message text', 'John', '10')
        request_body = {'request_items': dict(
            (tname, [{'put_request': {'item': item}}]) for tname in tnames)}

        key_conditions = {
            self.hashkey: {
                'attribute_value_list': [{'S': 'forum1'}],
                'comparison_operator': 'EQ'
            },
            self.rangekey: {
                'attribute_value_list': [{'S': 'subject'}],
                'comparison_operator': 'BEGINS_WITH'
            }
        }
        headers, body = self.client.batch_write_item(request_body)
        self.assertIn('unprocessed_items', body)
        self.assertEqual(body['unprocessed_items'], {})
        for tname in tnames:
            headers, body = self.client.query(tname,
                                              key_conditions=key_conditions,
                                              consistent_read=True)
            self.assertEqual(1, body['count'])
            self.assertIn(item, body['items'])

    @attr(type=['BWI-16'])
    def test_batch_write_put_s(self):
        self._create_test_table(self.build_x_attrs('S'),
                                self.tname,
                                self.smoke_schema,
                                wait_for_active=True)
        item = self.build_x_item('S', 'forum1', 'subject2',
                                 ('message', 'S', 'message text'))
        request_body = {'request_items': {self.tname: [{'put_request':
                                                        {'item': item}}]}}
        headers, body = self.client.batch_write_item(request_body)
        self.assertIn('unprocessed_items', body)
        self.assertEqual(body['unprocessed_items'], {})
        key_conditions = {
            self.hashkey: {
                'attribute_value_list': [{'S': 'forum1'}],
                'comparison_operator': 'EQ'
            }
        }
        headers, body = self.client.query(self.tname,
                                          key_conditions=key_conditions,
                                          consistent_read=True)
        self.assertEqual(item, body['items'][0])

    @attr(type=['BWI-50'])
    def test_batch_write_correct_tname_and_put_request(self):
        self._create_test_table(self.build_x_attrs('S'),
                                self.tname,
                                self.smoke_schema,
                                wait_for_active=True)
        item = self.build_x_item('S', 'forum1', 'subject2')
        request_body = {'request_items': {self.tname: [{'put_request':
                                                        {'item': item}}]}}
        headers, body = self.client.batch_write_item(request_body)
        self.assertIn('unprocessed_items', body)
        self.assertEqual(body['unprocessed_items'], {})
        key_conditions = {
            self.hashkey: {
                'attribute_value_list': [{'S': 'forum1'}],
                'comparison_operator': 'EQ'
            }
        }
        headers, body = self.client.query(self.tname,
                                          key_conditions=key_conditions,
                                          consistent_read=True)
        self.assertEqual(item, body['items'][0])



# Copyright 2013 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import collections
from threading import Lock

from magnetodb.common import exception
from magnetodb.storage import models
from magnetodb.storage.table_info_repo import TableInfo


class CassandraTableInfoRepository(object):
    SYSTEM_TABLE_TABLE_INFO = 'magnetodb.table_info'
    __field_list = ("schema", "internal_name", "status")

    def _save_table_info_to_cache(self, context, table_info):
        tenant_tables_cache = self.__table_info_cache.get(context.tenant)
        if tenant_tables_cache is None:
            tenant_tables_cache = {}
            self.__table_info_cache[context.tenant] = tenant_tables_cache

        # TODO(dukhlov):
        # It is temporary solution to set internal_keyspace
        # In future keyspace should be saved in metadata too
        from magnetodb.storage.driver.cassandra import USER_PREFIX
        table_info.internal_keyspace = USER_PREFIX + context.tenant

        tenant_tables_cache[table_info.name] = table_info

    def _get_table_info_from_cache(self, context, table_name):
        tenant_tables_cache = self.__table_info_cache.get(context.tenant)
        if tenant_tables_cache is None:
            return None
        return tenant_tables_cache.get(table_name)

    def _remove_table_schema_from_cache(self, context, table_name):
        tenant_tables_cache = self.__table_info_cache.get(context.tenant)
        if tenant_tables_cache is None:
            return None

        return tenant_tables_cache.pop(table_name, None)

    def __init__(self, cluster_handler):
        self.__cluster_handler = cluster_handler
        self.__table_info_cache = {}
        self.__table_cache_lock = Lock()

    def get(self, context, table_name, refresh_field_list=None):
        table_info = self._get_table_info_from_cache(context, table_name)
        if table_info:
            if refresh_field_list:
                self.refresh(context, table_info, refresh_field_list)
            return table_info

        with self.__table_cache_lock:
            table_info = self._get_table_info_from_cache(context, table_name)
            if table_info:
                return table_info

            table_info = TableInfo(table_name)
            self.refresh(context, table_info)
            self._save_table_info_to_cache(context, table_info)
        return table_info

    def get_tenant_table_names(self, context, exclusive_start_table_name=None,
                               limit=None):
        query_builder = collections.deque()
        query_builder.append(
            "SELECT name FROM {} WHERE tenant='{}'".format(
                self.SYSTEM_TABLE_TABLE_INFO, context.tenant
            )
        )

        if exclusive_start_table_name:
            query_builder.append(
                " AND name > '{}'".format(exclusive_start_table_name)
            )

        if limit:
            query_builder.append(" LIMIT {}".format(limit))

        tables = self.__cluster_handler.execute_query(
            "".join(query_builder), consistent=True
        )

        return [row['name'] for row in tables]

    def refresh(self, context, table_info, field_list=__field_list):
        query_builder = collections.deque()
        query_builder.append("SELECT ")
        query_builder.append(",".join(map('"{}"'.format, field_list)))

        query_builder.append(
            " FROM {} WHERE tenant='{}' AND name='{}'".format(
                self.SYSTEM_TABLE_TABLE_INFO,
                context.tenant, table_info.name
            )
        )

        result = self.__cluster_handler.execute_query(
            "".join(query_builder), consistent=True
        )

        if not result:
            self._remove_table_schema_from_cache(context, table_info.name)
            raise exception.TableNotExistsException(
                "Table '{}' does not exist".format(table_info.name)
            )
        for name, value in result[0].iteritems():
            if name == "schema":
                value = models.ModelBase.from_json(value)
            setattr(table_info, name, value)
        return True

    def update(self, context, table_info, field_list=None):
        if not field_list:
            field_list = self.__field_list

        query_builder = collections.deque()
        query_builder.append(
            "UPDATE {} SET ".format(self.SYSTEM_TABLE_TABLE_INFO)
        )

        query_builder.append(
            ",".join(
                [
                    '"{}"=\'{}\''.format(field, getattr(table_info, field))
                    for field in field_list
                ]
            )
        )

        query_builder.append(
            " WHERE tenant='{}' AND name='{}' IF exists=1".format(
                context.tenant, table_info.name
            )
        )

        result = self.__cluster_handler.execute_query(
            "".join(query_builder), consistent=True
        )

        if not result[0]['[applied]']:
            raise exception.TableNotExistsException(
                "Table {} is not exists".format(table_info.name)
            )
        return True

    def save(self, context, table_info):
        query_builder = collections.deque()
        query_builder.append(
            'INSERT INTO {} '
            '(exists, tenant, name, "schema", status, internal_name)'
            "VALUES(1, '{}', '{}'".format(
                self.SYSTEM_TABLE_TABLE_INFO, context.tenant, table_info.name
            )
        )

        if table_info.schema:
            query_builder.append(",'{}'".format(table_info.schema.to_json()))
        else:
            query_builder.append(",null")

        if table_info.status:
            query_builder.append(",'{}'".format(table_info.status))
        else:
            query_builder.append(",null")

        try:
            internal_name = table_info.internal_name
        except AttributeError:
            internal_name = None

        if internal_name:
            query_builder.append(",'{}'".format(table_info.internal_name))
        else:
            query_builder.append(",null")

        query_builder.append(") IF NOT EXISTS")

        result = self.__cluster_handler.execute_query(
            "".join(query_builder), consistent=True
        )

        if not result[0]['[applied]']:
            raise exception.TableAlreadyExistsException(
                "Table {} already exists".format(table_info.name)
            )

        self._save_table_info_to_cache(context, table_info)
        return True

    def delete(self, context, table_name):
        query = (
            "DELETE FROM {}"
            " WHERE tenant='{}' AND name='{}'".format(
                self.SYSTEM_TABLE_TABLE_INFO, context.tenant, table_name
            )
        )
        self.__cluster_handler.execute_query(query, consistent=True)

        return True
